import io

from cffi.api import FFI
from cffi.recompiler import Recompiler


def _execfile(pysrc, filename, globs: dict):
    compiled = compile(source=pysrc, filename=filename, mode="exec")
    exec(compiled, globs, globs)  # noqa: S102


def find_ffi_in_python_script(pysrc: str, filename: str, ffivar: str):
    globs = {"__name__": "gen-cffi-src"}
    _execfile(pysrc, filename, globs)
    if ffivar not in globs:
        raise NameError(f"Expected to find the FFI object with the name {ffivar!r}, but it was not found.")
    ffi = globs[ffivar]
    if not isinstance(ffi, FFI) and callable(ffi):
        # Maybe it's a callable that returns a FFI
        ffi = ffi()
    if not isinstance(ffi, FFI):
        raise TypeError(f"Found an object with the name {ffivar!r} but it was not an instance of cffi.api.FFI")
    # TODO: improve this; https://github.com/python-cffi/cffi/issues/47
    module_name, source, source_extension, kwds = ffi._assigned_source
    return module_name, source, ffi


def make_ffi_from_sources(modulename: str, cdef: str, csrc: str):
    ffibuilder = FFI()
    ffibuilder.cdef(cdef)
    ffibuilder.set_source(modulename, csrc)
    return ffibuilder


def generate_c_source(module_name: str, csrc: str, ffi: FFI):
    # TODO: improve this; https://github.com/python-cffi/cffi/issues/47
    recompiler = Recompiler(ffi, module_name)
    recompiler.collect_type_table()
    recompiler.collect_step_tables()
    output = io.StringIO()
    recompiler.write_c_source_to_f(output, csrc)
    return output.getvalue()



from cffi import FFI

ffibuilder = FFI()

ffibuilder.cdef(
    """\
int netstring_read(char *buffer, size_t buffer_length, char **netstring_start, size_t *netstring_length);

#define NETSTRING_ERROR_TOO_LONG     -1
#define NETSTRING_ERROR_NO_COLON     -2
#define NETSTRING_ERROR_TOO_SHORT    -3
#define NETSTRING_ERROR_NO_COMMA     -4
#define NETSTRING_ERROR_LEADING_ZERO -5
#define NETSTRING_ERROR_NO_LENGTH    -6
"""
)

ffibuilder.set_source(
    "netstring._netstring",
    """\
#include "netstring.h"
""",
)

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)



from ._netstring import ffi
from ._netstring import lib


class NetstringError(Exception):
    pass


error_codes = {
    lib.NETSTRING_ERROR_TOO_LONG: "Input buffer is too long",
    lib.NETSTRING_ERROR_NO_COLON: "No colon was found after the number",
    lib.NETSTRING_ERROR_TOO_SHORT: "Number of bytes greater than buffer length",
    lib.NETSTRING_ERROR_NO_COMMA: "No comma was found at the end",
    lib.NETSTRING_ERROR_LEADING_ZERO: "Leading zeros are not allowed",
    lib.NETSTRING_ERROR_NO_LENGTH: "Length not given at start of netstring",
}


def parse_netstring(somebytes: bytes) -> bytes:
    """Return a parsed netstring.

    >>> parse_netstring(b"12:hello world!,")
    b"hello world!"
    """
    buf = ffi.from_buffer(somebytes)
    netstring_start = ffi.new("char **")
    netstring_len = ffi.new("size_t *")
    result = lib.netstring_read(buf, len(somebytes), netstring_start, netstring_len)
    if result != 0:
        raise NetstringError(error_codes[result])
    return ffi.string(netstring_start[0], netstring_len[0])



import pytest

import cffi_buildtool.gen

SIMPLE = """\
from cffi import FFI

ffibuilder = FFI()

ffibuilder.cdef("int netstring_read(char *buffer, size_t buffer_length, char **netstring_start, size_t *netstring_length);")

ffibuilder.set_source("netstring._netstring", '#include "netstring.h"')

something_else = 42
"""

CALLABLE = """\
from cffi import FFI

def make_ffi():
    ffibuilder = FFI()
    ffibuilder.cdef("int netstring_read(char *buffer, size_t buffer_length, char **netstring_start, size_t *netstring_length);")
    ffibuilder.set_source("netstring._netstring", '#include "netstring.h"')
    return ffibuilder

def something_else():
    return 42
"""


def test_exec_python():
    module_name, csrc, ffi = cffi_buildtool.gen.find_ffi_in_python_script(SIMPLE, "_netstring.c.py", "ffibuilder")
    assert module_name == "netstring._netstring"
    assert csrc.strip() == '#include "netstring.h"'
    cdef = "\n".join(ffi._cdefsources)
    assert "int netstring_read" in cdef


def test_exec_python_callable():
    module_name, csrc, ffi = cffi_buildtool.gen.find_ffi_in_python_script(CALLABLE, "_netstring.c.py", "make_ffi")
    assert module_name == "netstring._netstring"
    assert csrc.strip() == '#include "netstring.h"'
    cdef = "\n".join(ffi._cdefsources)
    assert "int netstring_read" in cdef


def test_exec_python_not_found():
    with pytest.raises(NameError, match="Expected to find the FFI object with the name 'notfound', but it was not found."):
        cffi_buildtool.gen.find_ffi_in_python_script(SIMPLE, "_netstring.c.py", "notfound")


def test_exec_python_bad_type():
    with pytest.raises(TypeError, match=".+not an instance of cffi.api.FFI"):
        cffi_buildtool.gen.find_ffi_in_python_script(SIMPLE, "_netstring.c.py", "something_else")


def test_exec_python_callable_bad_type():
    with pytest.raises(TypeError, match=".+not an instance of cffi.api.FFI"):
        cffi_buildtool.gen.find_ffi_in_python_script(CALLABLE, "_netstring.c.py", "something_else")



import pathlib

import pytest

import cffi_buildtool.cli


def dont_exit(_status: int):
    pass


def test_read_sources(examples: pathlib.Path, tmp_path: pathlib.Path, monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(cffi_buildtool.cli.parser, "exit", dont_exit)
    srcdir = examples / "read-sources" / "cffi" / "netstring"
    cdef = srcdir / "_netstring.cdef.txt"
    csrc = srcdir / "_netstring.csrc.c"
    output = tmp_path / "out.c"
    cffi_buildtool.cli.run(["read-sources", "_netstring", str(cdef), str(csrc), str(output)])
    generated = output.read_text()
    assert "netstring_read" in generated


def test_exec_python(examples: pathlib.Path, tmp_path: pathlib.Path, monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(cffi_buildtool.cli.parser, "exit", dont_exit)
    pyfile = examples / "exec-python" / "cffi" / "netstring" / "_netstring.c.py"
    output = tmp_path / "out.c"
    cffi_buildtool.cli.run(["exec-python", str(pyfile), str(output)])
    generated = output.read_text()
    assert "netstring_read" in generated


def test_read_sources_same_input_fails(capsys: pytest.CaptureFixture):
    with pytest.raises(SystemExit):
        cffi_buildtool.cli.run(["read-sources", "_netstring", "-", "-", "-"])
    _stdout, stderr = capsys.readouterr()
    assert "are the same file and should not be" in stderr



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -mcffi_buildtool` python will execute
    ``__main__.py`` as a script. That means there will not be any
    ``cffi_buildtool.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there"s no ``cffi_buildtool.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""

import argparse
import typing

from .gen import find_ffi_in_python_script
from .gen import generate_c_source
from .gen import make_ffi_from_sources


def exec_python(*, output: typing.TextIO, pyfile: typing.TextIO, ffi_var: str):
    with pyfile:
        module_name, csrc, ffi = find_ffi_in_python_script(pyfile.read(), pyfile.name, ffi_var)
    generated = generate_c_source(module_name, csrc, ffi)
    with output:
        output.write(generated)


def read_sources(*, output: typing.TextIO, module_name: str, cdef_input: typing.TextIO, csrc_input: typing.TextIO):
    with csrc_input, cdef_input:
        csrc = csrc_input.read()
        cdef = cdef_input.read()
    ffi = make_ffi_from_sources(module_name, cdef, csrc)
    generated = generate_c_source(module_name, csrc, ffi)
    with output:
        output.write(generated)


parser = argparse.ArgumentParser(description="Command description.")
subparsers = parser.add_subparsers(dest="mode")

exec_python_parser = subparsers.add_parser("exec-python", help="Execute a Python script to build a FFI object")
exec_python_parser.add_argument(
    "--ffi-var", default="ffibuilder", help="Name of the FFI object in the Python script; defaults to 'ffibuilder'."
)
exec_python_parser.add_argument("pyfile", type=argparse.FileType("r", encoding="utf-8"), help="Path to the Python script")
exec_python_parser.add_argument("output", type=argparse.FileType("w", encoding="utf-8"), help="Output path for the C source")

read_sources_parser = subparsers.add_parser("read-sources", help="Read CDEF and C source prelude files")
read_sources_parser.add_argument("module_name", help="Full name of the generated module, including packages")
read_sources_parser.add_argument("cdef", type=argparse.FileType("r", encoding="utf-8"), help="File containing C definitions")
read_sources_parser.add_argument("csrc", type=argparse.FileType("r", encoding="utf-8"), help="File containing C source prelude")
read_sources_parser.add_argument("output", type=argparse.FileType("w", encoding="utf-8"), help="Output path for the C source")


def run(args: typing.Optional[typing.Sequence[str]] = None):
    args = parser.parse_args(args=args)
    if args.mode == "exec-python":
        exec_python(output=args.output, pyfile=args.pyfile, ffi_var=args.ffi_var)
    elif args.mode == "read-sources":
        if args.cdef is args.csrc:
            parser.error("--cdef and --csrc are the same file and should not be")
        read_sources(output=args.output, module_name=args.module_name, cdef_input=args.cdef, csrc_input=args.csrc)
    else:  # pragma: no cover
        raise AssertionError("Unknown subcommand; should not ever happen.")
    parser.exit(0)
